#include <iostream>
#include "CataCamera.h"

CataCamera::CataCamera(double xi, double k1, double k2, double p1, double p2,
		       double gamma1, double gamma2, double u0, double v0) {
  //cout << "Creating class CataCamera." << endl;
  cout << "xi = " << xi << 
    ", k1 = "<< k1 << ", k2 = " << k2 << ", p1 = " << p1 << ", p2 = " << p2 <<
    ", gamma1 = " << gamma1 << ", gamma2 = " << gamma2 << 
    ", u0 = " << u0 << ", v0 = " << v0 << endl;
  
  this->xi = xi;
  if ((k1==0)&&(k2==0)&&(p1==0)&&(p2==0)) {
    cout << "No distortion!!!" << endl;
    this->no_distortion = 1;
  } else {
    this->no_distortion = 0;
  }
  // !!!!!!!!  
  //this->no_distortion = 1;
  // !!!!!!!!  
  this->k1 = k1;
  this->k2 = k2;
  this->p1 = p1;
  this->p2 = p2;
  this->gamma1 = gamma1;
  this->gamma2 = gamma2;
  this->u0 = u0;
  this->v0 = v0;
  // Inverse camera projection matrix parameters
  this->inv_K11 = 1/gamma1;
  this->inv_K13 = -u0/gamma1;
  this->inv_K22 = 1/gamma2;
  this->inv_K23 = -v0/gamma2;
}

CataCamera::CataCamera(char *fileName) {
  FILE * fid = fopen (fileName, "r");

  if (fid==NULL) {
    fprintf(stderr,"Cannot open file '%s'.\n",fileName);
    return;
  }

  double dump;

  fread(&xi, sizeof(double), 1, fid);
  
  fread(&k1, sizeof(double), 1, fid);
  fread(&k2, sizeof(double), 1, fid);
  fread(&p1, sizeof(double), 1, fid);
  fread(&p2, sizeof(double), 1, fid);

  fread(&dump, sizeof(double), 1, fid);

  if (dump!=0) {
    printf("WARNING : we do not currently take into account the 'k5' parameter in this C++ class.\n");
    printf("Assuming 'k5=0'.\n");
  }

  fread(&gamma1, sizeof(double), 1, fid);
  fread(&gamma2, sizeof(double), 1, fid);
  fread(&u0, sizeof(double), 1, fid);
  fread(&v0, sizeof(double), 1, fid);
  fread(&dump, sizeof(double), 1, fid);

  if (dump!=0) {
    printf("WARNING : we do not currently take into account the 'alpha_c' (skew : 's') parameter in this C++ class.\n");
    printf("Assuming 'alpha_c=0'.\n");
  }
    
  fread(&roi_min_x, sizeof(double), 1, fid);
  fread(&roi_min_y, sizeof(double), 1, fid);
  fread(&roi_max_x, sizeof(double), 1, fid);
  fread(&roi_max_y, sizeof(double), 1, fid);

  fclose(fid);

   // Inverse camera projection matrix parameters
  this->inv_K11 = 1/gamma1;
  this->inv_K13 = -u0/gamma1;
  this->inv_K22 = 1/gamma2;
  this->inv_K23 = -v0/gamma2;

  if ((k1==0)&&(k2==0)&&(p1==0)&&(p2==0)) {
    cout << "No distortion!!!" << endl;
    this->no_distortion = 1;
  } else {
    this->no_distortion = 0;
  }

  cout << "xi = " << xi << 
    ", k1 = "<< k1 << ", k2 = " << k2 << ", p1 = " << p1 << ", p2 = " << p2 <<
    ", gamma1 = " << gamma1 << ", gamma2 = " << gamma2 << 
    ", u0 = " << u0 << ", v0 = " << v0 << 
    ", roi_min = (" << roi_min_x << "," << roi_min_y << ")" << 
    ", roi_max = (" << roi_max_x << "," << roi_max_y << ")" << endl;

}


CataCamera::~CataCamera() {
  cout << "Destroying class CataCamera!" << endl;
}

/** 
 * \brief Lifts a point from the image plane to the unit sphere
 *
 * \param u u image coordinate
 * \param v v image coordinate
 * \param X X coordinate of the point on the sphere
 * \param Y Y coordinate of the point on the sphere
 * \param Z Z coordinate of the point on the sphere
 */
void CataCamera::lift_sphere(double u, double v, double *X, double *Y, double *Z) {
  double mx_d, my_d,mx2_d, mxy_d, my2_d, mx_u, my_u;
  double rho2_d, rho4_d, radDist_d, Dx_d, Dy_d, inv_denom_d;
  double lambda;
  
  // Lift points to normalised plane
  // Matlab points start at 1 (calibration)
  mx_d = inv_K11*(u)+inv_K13;
  my_d = inv_K22*(v)+inv_K23;
   
  if (no_distortion) {
    mx_u = mx_d;
    my_u = my_d;
  } else {
    // Apply inverse distortion model
    if (0) {
      // Inverse distortion model
      // proposed by Heikkila
      mx2_d = mx_d*mx_d;
      my2_d = my_d*my_d;
      mxy_d = mx_d*my_d;
      rho2_d = mx2_d+my2_d;
      rho4_d = rho2_d*rho2_d;
      radDist_d = k1*rho2_d+k2*rho4_d;
      Dx_d = mx_d*radDist_d + p2*(rho2_d+2*mx2_d) + 2*p1*mxy_d;
      Dy_d = my_d*radDist_d + p1*(rho2_d+2*my2_d) + 2*p2*mxy_d;
      inv_denom_d = 1/(1+4*k1*rho2_d+6*k2*rho4_d+8*p1*my_d+8*p2*mx_d);
  
      mx_u = mx_d - inv_denom_d*Dx_d;
      my_u = my_d - inv_denom_d*Dy_d;
    } else {
      // Recursive distortion model
      int n = 6;
      double dx_u, dy_u;
      distortion(mx_d,my_d,&dx_u,&dy_u);
      // Approximate value
      mx_u = mx_d-dx_u;
      my_u = my_d-dy_u;

      for(int i=1;i<n;i++) {
	distortion(mx_u,my_u,&dx_u,&dy_u);
	mx_u = mx_d-dx_u;
	my_u = my_d-dy_u;
      }
    }
  }

  // Lift normalised points to the sphere (inv_hslash)
  if (xi==1) {
    lambda = 2/(mx_u*mx_u+my_u*my_u+1);
    *X = lambda*mx_u;
    *Y = lambda*my_u;
    *Z = lambda-1;
  } else {
    //printf("!!!!! Here !!!!!");
    lambda = (xi+sqrt(1+(1-xi*xi)*(mx_u*mx_u+my_u*my_u)))/(1+mx_u*mx_u+my_u*my_u);
    *X = lambda*mx_u;
    *Y = lambda*my_u;
    *Z = lambda-xi;
  }
}

/** 
 * \brief Lifts a point from the image plane to its projective ray
 *
 * \param u u image coordinate
 * \param v v image coordinate
 * \param X X coordinate of the projective ray
 * \param Y Y coordinate of the projective ray
 * \param Z Z coordinate of the projective ray
 */
void CataCamera::lift_projective(double u, double v, double *X, double *Y, double *Z) {
  double mx_d, my_d,mx2_d, mxy_d, my2_d, mx_u, my_u;
  double rho2_d, rho4_d, radDist_d, Dx_d, Dy_d, inv_denom_d;
  double lambda;
  
  // Lift points to normalised plane
  // Matlab points start at 1 (calibration)
  mx_d = inv_K11*(u)+inv_K13;
  my_d = inv_K22*(v)+inv_K23;

  if (no_distortion) {
    mx_u = mx_d;
    my_u = my_d;
  } else {
    if (0) {
      // Apply inverse distortion model 
      // proposed by Heikkila
      mx2_d = mx_d*mx_d;
      my2_d = my_d*my_d;
      mxy_d = mx_d*my_d;
      rho2_d = mx2_d+my2_d;
      rho4_d = rho2_d*rho2_d;
      radDist_d = k1*rho2_d+k2*rho4_d;
      Dx_d = mx_d*radDist_d + p2*(rho2_d+2*mx2_d) + 2*p1*mxy_d;
      Dy_d = my_d*radDist_d + p1*(rho2_d+2*my2_d) + 2*p2*mxy_d;
      inv_denom_d = 1/(1+4*k1*rho2_d+6*k2*rho4_d+8*p1*my_d+8*p2*mx_d);
      
      mx_u = mx_d - inv_denom_d*Dx_d;
      my_u = my_d - inv_denom_d*Dy_d;
    } else {
      // Recursive distortion model
      int n = 8;
      double dx_u, dy_u;
      distortion(mx_d,my_d,&dx_u,&dy_u);
      // Approximate value
      mx_u = mx_d-dx_u;
      my_u = my_d-dy_u;
      
      for(int i=1;i<n;i++) {
	distortion(mx_u,my_u,&dx_u,&dy_u);
	mx_u = mx_d-dx_u;
	my_u = my_d-dy_u;
      }
    }
  }
 
  // Obtain a projective ray
  if (xi==1) {
    *X = mx_u;
    *Y = my_u;
    *Z = (1-mx_u*mx_u-my_u*my_u)/2;
  } else {
    // Reuse variable
    rho2_d = mx_u*mx_u+my_u*my_u;
    *X = mx_u;
    *Y = my_u;
    *Z = 1-xi*(rho2_d+1)/(xi+sqrt(1+(1-xi*xi)*rho2_d));
  }
}


/** 
 * \brief Project a 3D points (\a x,\a y,\a z) to the image plane in (\a u,\a v)
 *
 * \param x 3D point x coordinate
 * \param y 3D point y coordinate
 * \param z 3D point z coordinate
 * \param u return value, contains the image point u coordinate
 * \param v return value, contains the image point v coordinate
 */
void CataCamera::space2plane(double x, double y, double z, double *u, double *v) {
  double mx_u, my_u, mx_d, my_d;

  // Project points to the normalised plane
  z = z+xi*sqrt(x*x+y*y+z*z);
  mx_u = x/z;
  my_u = y/z;
  
  if (no_distortion) {
    mx_d = mx_u;
    my_d = my_u;
  } else {
    // Apply distortion
    double dx_u, dy_u;
    distortion(mx_u,my_u,&dx_u,&dy_u);
    mx_d = mx_u+dx_u;
    my_d = my_u+dy_u;
  }
  
  // Apply generalised projection matrix
  // Matlab points start at 1
  *u = gamma1*mx_d+u0;
  *v = gamma2*my_d+v0;
}


/** 
 * \brief Project a 3D points (\a x,\a y,\a z) to the image plane in (\a u,\a v)
 *        and calculate jacobian
 *
 * \param x 3D point x coordinate
 * \param y 3D point y coordinate
 * \param z 3D point z coordinate
 * \param u return value, contains the image point u coordinate
 * \param v return value, contains the image point v coordinate
 */
void CataCamera::space2plane(double x, double y, double z, 
			     double *u, double *v,
			     double *dudx, double *dvdx,
			     double *dudy, double *dvdy,
			     double *dudz, double *dvdz) {
  double mx_u, my_u, mx_d, my_d;
  double norm,inv_denom;
  double dxdmx, dydmx, dxdmy, dydmy;

  norm = sqrt(x*x+y*y+z*z);
  // Project points to the normalised plane
  inv_denom = 1/(z+xi*norm);
  mx_u = inv_denom*x;
  my_u = inv_denom*y;

  // Calculate jacobian
  inv_denom = inv_denom*inv_denom/norm;
  *dudx = inv_denom*(norm*z+xi*(y*y+z*z));
  *dvdx = -inv_denom*xi*x*y;
  *dudy = *dvdx;
  *dvdy = inv_denom*(norm*z+xi*(x*x+z*z));
  inv_denom = inv_denom*(-xi*z-norm); // reuse variable
  *dudz = x*inv_denom;
  *dvdz = y*inv_denom;

  if (no_distortion) {
    mx_d = mx_u;
    my_d = my_u;
  } else {
    // Apply distortion
    double dx_u, dy_u;
    distortion(mx_u,my_u,&dx_u,&dy_u,&dxdmx,&dydmx,&dxdmy,&dydmy);
    mx_d = mx_u+dx_u;
    my_d = my_u+dy_u;
  }

  // Make the product of the jacobians
  // and add projection matrix jacobian
  inv_denom = gamma1*(*dudx*dxdmx+*dvdx*dxdmy); // reuse
  *dvdx = gamma2*(*dudx*dydmx+*dvdx*dydmy);
  *dudx = inv_denom;
  
  inv_denom = gamma1*(*dudy*dxdmx+*dvdy*dxdmy); // reuse
  *dvdy = gamma2*(*dudy*dydmx+*dvdy*dydmy);
  *dudy = inv_denom;
    
  inv_denom = gamma1*(*dudz*dxdmx+*dvdz*dxdmy); // reuse
  *dvdz = gamma2*(*dudz*dydmx+*dvdz*dydmy);
  *dudz = inv_denom;  

  // Apply generalised projection matrix
  // Matlab points start at 1
  *u = gamma1*mx_d+u0;
  *v = gamma2*my_d+v0;
}





/** 
 * \brief Projects an undistorted 2D point (\a mx_u,\a my_u) to the image plane in (\a u,\a v)
 *
 * \param mx_u 2D point x coordinate
 * \param my_u 3D point y coordinate
 * \param u return value, contains the image point u coordinate
 * \param v return value, contains the image point v coordinate
 */
void CataCamera::undist2plane(double mx_u, double my_u, double *u, double *v) {
  double mx_d, my_d;

  if (no_distortion) {
    mx_d = mx_u;
    my_d = my_u;
  } else {
    // Apply distortion
    double dx_u, dy_u;
    distortion(mx_u,my_u,&dx_u,&dy_u);
    mx_d = mx_u+dx_u;
    my_d = my_u+dy_u;
  }
  
  // Apply generalised projection matrix
  // Matlab points start at 1
  *u = gamma1*mx_d+u0;
  *v = gamma2*my_d+v0;
}

/** 
 * \brief Calculates the reprojection distance between points
 *
 * \param x1 first 3D point x coordinate
 * \param y1 first 3D point y coordinate
 * \param z1 first 3D point z coordinate
 * \param x2 second 3D point x coordinate
 * \param y2 second 3D point y coordinate
 * \param z2 second 3D point z coordinate
 * \param dist return value, euclidean distance in the plane
 */
void CataCamera::reprojectionDist(double x1, double y1, double z1, double x2, double y2, double z2, double *dist) {
  double u1,v1,u2,v2;

  space2plane(x1,y1,z1,&u1,&v1);
  space2plane(x2,y2,z2,&u2,&v2);

  *dist = sqrt((u1-u2)*(u1-u2)+(v1-v2)*(v1-v2));
}

/** 
 * \brief Apply distortion to input point (from the normalised plane)
 *  
 * \param mx_u undistorted x coordinate of point on the normalised plane
 * \param my_u undistorted y coordinate of point on the normalised plane
 * \param dx return value, to obtain the distorted point : mx_d = mx_u+dx_u
 * \param dy return value, to obtain the distorted point : my_d = my_u+dy_u
 */
void CataCamera::distortion(double mx_u, double my_u, 
			   double *dx_u, double *dy_u) {
  double mx2_u, my2_u, mxy_u, rho2_u, rad_dist_u;

  mx2_u = mx_u*mx_u;
  my2_u = my_u*my_u;
  mxy_u = mx_u*my_u;
  rho2_u = mx2_u+my2_u;
  rad_dist_u = k1*rho2_u+k2*rho2_u*rho2_u;
  *dx_u = mx_u*rad_dist_u + 2*p1*mxy_u + p2*(rho2_u+2*mx2_u);
  *dy_u = my_u*rad_dist_u + 2*p2*mxy_u + p1*(rho2_u+2*my2_u);  
}

/** 
 * \brief Apply distortion to input point (from the normalised plane)
 *        and calculate jacobian
 *
 * \param mx_u undistorted x coordinate of point on the normalised plane
 * \param my_u undistorted y coordinate of point on the normalised plane
 * \param dx return value, to obtain the distorted point : mx_d = mx_u+dx_u
 * \param dy return value, to obtain the distorted point : my_d = my_u+dy_u
 */
void CataCamera::distortion(double mx_u, double my_u, 
			    double *dx_u, double *dy_u,
			    double *dxdmx, double *dydmx,
			    double *dxdmy, double *dydmy) {
  double mx2_u, my2_u, mxy_u, rho2_u, rad_dist_u;

  mx2_u = mx_u*mx_u;
  my2_u = my_u*my_u;
  mxy_u = mx_u*my_u;
  rho2_u = mx2_u+my2_u;
  rad_dist_u = k1*rho2_u+k2*rho2_u*rho2_u;
  *dx_u = mx_u*rad_dist_u + 2*p1*mxy_u + p2*(rho2_u+2*mx2_u);
  *dy_u = my_u*rad_dist_u + 2*p2*mxy_u + p1*(rho2_u+2*my2_u);  

  *dxdmx = 1 + rad_dist_u + k1*2*mx2_u + k2*rho2_u*4*mx2_u + 2*p1*my_u + 6*p2*mx_u;
  *dydmx = k1*2*mx_u*my_u + k2*4*rho2_u*mx_u*my_u + p1*2*mx_u + 2*p2*my_u;
  *dxdmy = *dydmx;
  *dydmy = 1 + rad_dist_u + k1*2*my2_u + k2*rho2_u*4*my2_u + 6*p1*my_u + 2*p2*mx_u;  
}

void CataCamera::getV(double *V) {
  V[0] = xi;
  V[1] = k1;
  V[2] = k2;
  V[3] = p1;
  V[4] = p2;
  V[5] = 0; // k5
  V[6] = 0; // alpha
  V[7] = gamma1;
  V[8] = gamma2;
  V[9] = u0;
  V[10] = v0;
}
