#include <iostream.h>
#include <math.h>
#include "mex.h"


class CataCamera {
 public:
  /**
   * \brief Constructor from the projection model parameters
   */
  CataCamera(double xi, double k1, double k2, double p1, double p2,
	    double gamma1, double gamma2, double u0, double v0);
  /**
   * \brief Constructor from a file containing the projection model parameters
   */
  CataCamera(char * fileName);
    
  // Lift points from the image plane to the sphere
  void lift_sphere(double u, double v, double *X, double *Y, double *Z);
  //%output X
  //%output Y
  //%output Z

  // Lift points from the image plane to the projective space
  void lift_projective(double u, double v, double *X, double *Y, double *Z);
  //%output X
  //%output Y
  //%output Z

  // Projects 3D points to the image plane (Pi function)
  void space2plane(double x, double y, double z, double *u, double *v);
  //%output u
  //%output v

  // Projects 3D points to the image plane (Pi function)
  // and calculates jacobian
  void space2plane(double x, double y, double z,
		   double *u, double *v,
		   double *dudx, double *dvdx,
		   double *dudy, double *dvdy,
		   double *dudz, double *dvdz);
  //%output u
  //%output v
  //%output dudx
  //%output dvdx
  //%output dudy
  //%output dvdy
  //%output dudz
  //%output dvdz


  void undist2plane(double mx_u, double my_u, double *u, double *v);
  //%output u
  //%output v

  void reprojectionDist(double x1, double y1, double z1, double x2, double y2, double z2, double *dist);
  //%output dist

  void distortion(double mx_u, double my_u, double *mx_d, double *my_d);
  void distortion(double mx_u, double my_u, double *mx_d, double *my_d,
		  double *dxdmx, double *dydmx,
		  double *dxdmy, double *dydmy);

  // V = (xi,k1,k2,k3,k4,k5,
  //      alpha,gamma1,gamma2,u0,v0)
  void getV(double *V);
  //%output V(1,11)

  ~CataCamera();
  
 private:
  double xi, k1, k2, p1, p2, gamma1, gamma2, u0, v0;
  double roi_min_x, roi_min_y, roi_max_x, roi_max_y;
  double inv_K11, inv_K13, inv_K22, inv_K23;
  bool no_distortion;
};
